@@ -20,7 +20,7 @@ import torch
 from matplotlib import pyplot as plt
 
 from slisemap.slisemap import Slisemap
-from slisemap.utils import _tonp, global_model
+from slisemap.utils import dict_array, dict_concat, global_model, tonp
 
 
 def _size(n: int, part: Union[float, int]) -> int:
@@ -39,7 +39,7 @@ def _frac(n: int, part: Union[float, int]) -> float:
 
 def global_model_losses(
     sm: Slisemap, indices: Optional[np.ndarray] = None, **kwargs
-) -> Tuple[torch.Tensor, torch.Tensor]:
+) -> torch.Tensor:
     """Train a global model
 
     Args:
@@ -48,7 +48,7 @@ def global_model_losses(
         **kwargs: Optional keyword arguments to LBFGS.
 
     Returns:
-        Tuple[torch.Tensor, torch.Tensor]: Tuple with vector of individual losses and vector of model coefficients.
+        torch.Tensor: Vector of individual losses for a global model.
     """
     if indices is None:
         X = sm.X
@@ -115,19 +115,17 @@ def plot_diagnostics(
     Returns:
         Optional[sns.FacetGrid]: Seaborn FacetGrid if show=False.
     """
-    import pandas as pd
-
     if isinstance(Z, Slisemap):
         Z = Z.get_Z(rotate=True)
     if len(diagnostics) == 1:
         for name, mask in diagnostics.items():
-            df = pd.DataFrame({"Z1": Z[:, 0], "Z2": Z[:, 1], name: mask})
+            df = dict_array({"Z1": Z[:, 0], "Z2": Z[:, 1], name: mask})
             g = sns.relplot(
                 data=df, x="Z1", y="Z2", hue=name, style=name, kind="scatter", **kwargs
             )
     elif summary:
         issues = reduce(lambda a, b: a + b.astype(int), diagnostics.values())
-        df = pd.DataFrame(
+        df = dict_array(
             {"Z1": Z[:, 0], "Z2": Z[:, 1], "Problem": issues > 0, "Severity": issues}
         )
         g = sns.relplot(
@@ -147,18 +145,14 @@ def plot_diagnostics(
                 kwargs["col_wrap"] = 3
             else:
                 kwargs["col_wrap"] = 4
-        df = pd.concat(
-            [
-                pd.DataFrame(
-                    {
-                        "Z1": Z[:, 0],
-                        "Z2": Z[:, 1],
-                        "Problem": mask,
-                        "Diagnostic": f"{diag} ({np.mean(mask) * 100:.1f} %)",
-                    }
-                )
-                for diag, mask in diagnostics.items()
-            ]
+        df = dict_concat(
+            {
+                "Z1": Z[:, 0],
+                "Z2": Z[:, 1],
+                "Problem": mask,
+                "Diagnostic": f"{diag} ({np.mean(mask) * 100:.1f} %)",
+            }
+            for diag, mask in diagnostics.items()
         )
         g = sns.relplot(
             data=df,
@@ -202,7 +196,7 @@ def heavyweight_diagnostic(
     Returns:
         np.ndarray: Boolean mask of problematic data items.
     """
-    return _tonp(sm.get_W(numpy=False).diag() > _frac(sm.n, min_size))
+    return tonp(sm.get_W(numpy=False).diag() > _frac(sm.n, min_size))
 
 
 def lightweight_diagnostic(
@@ -217,7 +211,7 @@ def lightweight_diagnostic(
     Returns:
         np.ndarray: Boolean mask of problematic data items.
     """
-    return _tonp(sm.get_W(numpy=False).diag() < (1 / _size(sm.n, max_size)))
+    return tonp(sm.get_W(numpy=False).diag() < (1 / _size(sm.n, max_size)))
 
 
 def weight_neighbourhood_diagnostic(
@@ -235,7 +229,7 @@ def weight_neighbourhood_diagnostic(
     """
     min_size = _size(sm.n, min_size)
     max_size = _size(sm.n, max_size)
-    return _tonp((sm.get_W(numpy=False) > (1 / max_size)).sum(1) < min_size)
+    return tonp((sm.get_W(numpy=False) > (1 / max_size)).sum(1) < min_size)
 
 
 def loss_neighbourhood_diagnostic(
@@ -296,7 +290,22 @@ def global_loss_diagnostic(
     ]
     treshold = np.mean(glosses) + np.std(glosses) * sd
     llosses = sm.get_L(numpy=False).sum(1)
-    return (llosses < treshold).cpu().numpy()
+    return tonp(llosses < treshold)
+
+
+def quantile_loss_diagnostic(sm: Slisemap, quantile: float = 0.4) -> np.ndarray:
+    """Check if any fidelity is worse than a quantile of all losses.
+
+    Args:
+        sm (Slisemap): Trained Slisemap solution.
+        quantile (float, optional): The quantile percentage. Defaults to 0.4.
+
+    Returns:
+        np.ndarray: Boolean mask of problematic data items.
+    """
+    L = sm.get_L(numpy=False)
+    treshold = torch.quantile(L.ravel(), _frac(sm.n, quantile))
+    return tonp(L.diag() > treshold)
 
 
 def optics_diagnostic(sm: Slisemap, min_size: Union[float, int] = 0.1, **kwargs):
@@ -342,6 +351,7 @@ def diagnose(
             "Weight Neighbourhood": weight_neighbourhood_diagnostic(
                 sm, min_size, max_size
             ),
+            "Quantile Loss": quantile_loss_diagnostic(sm, max_size),
         }
     else:
         return {
@@ -354,4 +364,5 @@ def diagnose(
             "Loss Neighbourhood": loss_neighbourhood_diagnostic(sm, min_size),
             "Global Loss": global_loss_diagnostic(sm),
             "Clustering": optics_diagnostic(sm, min_size),
+            "Quantile Loss": quantile_loss_diagnostic(sm, max_size),
         }
