@@ -13,6 +13,7 @@ import torch
 from matplotlib.figure import Figure
 from matplotlib.lines import Line2D
 from matplotlib.patches import Patch
+from sklearn.cluster import KMeans
 
 from slisemap.escape import escape_neighbourhood
 from slisemap.local_models import linear_regression, linear_regression_loss
@@ -22,10 +23,10 @@ from slisemap.utils import (
     CheckConvergence,
     PCA_rotation,
     _assert,
-    _tonp,
     _warn,
+    dict_concat,
     global_model,
-    varimax,
+    tonp,
 )
 
 
@@ -156,13 +157,13 @@ class Slisemap:
             self._Y = self._Y[:, None]
 
         if Z0 is None:
-            self._Z0 = self._X @ PCA_rotation(self._X, min(m, d))
-            if d > m:
+            self._Z0 = self._X @ PCA_rotation(self._X, d)
+            if self._Z0.shape[1] < d:
                 _warn(
                     "The number of embedding dimensions is larger than the number of data dimensions",
                     Slisemap,
                 )
-                Z0fill = torch.zeros(size=[n, d - m], **tensorargs)
+                Z0fill = torch.zeros(size=[n, d - self._Z0.shape[1]], **tensorargs)
                 self._Z0 = torch.cat((self._Z0, Z0fill), 1)
         else:
             self._Z0 = torch.as_tensor(Z0, **tensorargs)
@@ -189,14 +190,13 @@ class Slisemap:
                 lasso=self.lasso,
                 ridge=self.ridge,
             )
-            if torch.all(torch.isfinite(B0)):
-                self._B0 = B0.expand((n, coefficients))
-            else:
+            if not torch.all(torch.isfinite(B0)):
                 _warn(
                     "Optimising a global model as initialisation resulted in non-finite values. Consider using stronger regularisation (increase ``lasso`` or ``ridge``).",
                     Slisemap,
                 )
-                self._B0 = torch.zeros((n, coefficients), **tensorargs)
+                B0 = torch.zeros_like(B0)
+            self._B0 = B0.expand((n, coefficients))
         else:
             self._B0 = torch.as_tensor(B0, **tensorargs)
             if coefficients is None:
@@ -536,7 +536,7 @@ class Slisemap:
         else:
             Z = self._Z
         if rotate:
-            Z = varimax(Z)
+            Z = Z @ PCA_rotation(Z)
         if numpy:
             return Z.cpu().detach().numpy()
         else:
@@ -551,7 +551,7 @@ class Slisemap:
         Returns:
             Union[np.ndarray, torch.Tensor]: The B matrix
         """
-        return _tonp(self._B) if numpy else self._B.detach().clone()
+        return tonp(self._B) if numpy else self._B.detach().clone()
 
     def get_D(self, numpy: bool = True) -> Union[np.ndarray, torch.Tensor]:
         """Get the embedding distance matrix
@@ -564,7 +564,7 @@ class Slisemap:
         """
         Z = self.get_Z(rotate=False, numpy=False)
         D = self._distance(Z, Z)
-        return _tonp(D) if numpy else D.detach()
+        return tonp(D) if numpy else D.detach()
 
     def get_L(
         self,
@@ -585,7 +585,7 @@ class Slisemap:
         X = self._as_new_X(X)
         Y = self._as_new_Y(Y, X.shape[0])
         L = self.local_loss(self.local_model(X, self._B), Y, self._B)
-        return _tonp(L) if numpy else L.detach()
+        return tonp(L) if numpy else L.detach()
 
     def get_W(self, numpy: bool = True) -> Union[np.ndarray, torch.Tensor]:
         """Get the weight matrix
@@ -597,7 +597,7 @@ class Slisemap:
             Union[np.ndarray, torch.Tensor]: The W matrix
         """
         W = self.kernel(self.get_D(numpy=False))
-        return _tonp(W) if numpy else W.detach()
+        return tonp(W) if numpy else W.detach()
 
     def get_X(
         self, numpy: bool = True, intercept: bool = True
@@ -612,7 +612,7 @@ class Slisemap:
             Union[np.ndarray, torch.Tensor]: The X matrix
         """
         X = self._X if intercept or not self._intercept else self._X[:, :-1]
-        return _tonp(X) if numpy else X.detach().clone()
+        return tonp(X) if numpy else X.detach().clone()
 
     def get_Y(
         self, numpy: bool = True, ravel: bool = False
@@ -627,7 +627,7 @@ class Slisemap:
             Union[np.ndarray, torch.Tensor]: The Y matrix
         """
         Y = self._Y.ravel() if ravel else self._Y
-        return _tonp(Y) if numpy else Y.detach().clone()
+        return tonp(Y) if numpy else Y.detach().clone()
 
     def value(
         self, individual: bool = False, numpy: bool = True
@@ -643,7 +643,7 @@ class Slisemap:
         loss = self.get_loss_fn(individual)
         loss = loss(X=self._X, Y=self._Y, B=self._B, Z=self._Z)
         if individual:
-            return _tonp(loss) if numpy else loss.detach()
+            return tonp(loss) if numpy else loss.detach()
         else:
             return loss.cpu().detach().item() if numpy else loss.detach()
 
@@ -665,7 +665,7 @@ class Slisemap:
             entropy = (entropy.mean().exp() / self.n).detach()
             return entropy.cpu().item() if numpy else entropy
         else:
-            return _tonp(entropy) if numpy else entropy.detach()
+            return tonp(entropy) if numpy else entropy.detach()
 
     def lbfgs(self, max_iter: int = 500, *, only_B: bool = False, **kwargs) -> float:
 
@@ -712,6 +712,7 @@ class Slisemap:
         self,
         force_move: bool = True,
         escape_fn: Callable = escape_neighbourhood,
+        noise: float = 0.0,
     ):
         """Try to escape a local optimum by moving the items (embedding and local model) to the neighbourhoods best suited for them.
         This is done by finding another item (in the optimal neighbourhood) and copying its values for Z and B.
@@ -719,6 +720,7 @@ class Slisemap:
         Args:
             force_move (bool, optional): Do not allow the items to pair with themselves. Defaults to True.
             escape_fn (Callable, optional): Escape function (escape_neighbourhood/escape_greedy/escape_marginal). Defaults to escape_neighbourhood.
+            noise (float, optional): Scale of the noise added to the embedding matrix if it looses rank after an escape (recommended for gradient based optimisers). Defaults to 0.0.
         """
         self._B, self._Z = escape_fn(
             X=self._X,
@@ -733,6 +735,10 @@ class Slisemap:
             force_move=force_move,
             jit=self.jit,
         )
+        if noise > 0.0:
+            rank = torch.linalg.matrix_rank(self.Z - torch.mean(self.Z, 0, True))
+            if rank.item() < min(*self.Z.shape):
+                self._Z = torch.normal(self.Z, noise, generator=self._random_state)
         self._normalise()
 
     def _normalise(self):
@@ -742,23 +748,25 @@ class Slisemap:
 
     def optimise(
         self,
-        patience: int = 3,
+        patience: int = 2,
         max_escapes: int = 100,
         max_iter: int = 500,
         escape_fn: Callable = escape_neighbourhood,
         verbose: bool = False,
         noise: float = 1e-4,
+        only_B: bool = False,
         **kwargs,
     ) -> float:
         """Optimise Slisemap by alternating between Slisemap.lbfgs and Slisemap.escape until convergence.
 
         Args:
-            patience (int, optional): Number of escapes without improvement before stopping. Defaults to 3.
+            patience (int, optional): Number of escapes without improvement before stopping. Defaults to 2.
             max_escapes (int, optional): Maximum number of escapes. Defaults to 100.
             max_iter (int, optional): Maximum number of LBFGS iterations per round. Defaults to 500.
             escape_fn (Callable, optional): Escape function (escape_neighbourhood/escape_greedy/escape_marginal). Defaults to escape_neighbourhood.
             verbose (bool, optional): Print status messages. Defaults to False.
-            noise (float, optional): Scale of the noise used to avoid disappearing dimensions in the embedding after an escape (gradient based optimisers cannot recover them). Defaults to 1e-4.
+            noise (float, optional): Scale of the noise added to the embedding matrix if it looses rank after an escape. Defaults to 1e-4.
+            only_B (bool, optional): Only optimise the local models, not the embedding. Defaults to False.
             **kwargs: Optional keyword arguments to Slisemap.lbfgs.
 
         Returns:
@@ -767,24 +775,26 @@ class Slisemap:
         loss = np.repeat(np.inf, 2)
         loss[0] = self.lbfgs(max_iter=max_iter, only_B=True, **kwargs)
         if verbose:
-            print(f"LBFGS   0: {loss[0]:.2f}")
-        cc = CheckConvergence(patience)
-        for i in range(max_escapes):
-            self.escape(escape_fn=escape_fn)
+            i = 0
+            print(f"LBFGS  {i:2d}: {loss[0]:.2f}")
+        if only_B:
+            return loss[0]
+        cc = CheckConvergence(patience, max_escapes)
+        while not cc.has_converged(loss, self.copy):
+            self.escape(escape_fn=escape_fn, noise=noise)
             loss[1] = self.value()
             if verbose:
                 print(f"Escape {i:2d}: {loss[1]:.2f}")
-            if noise > 0.0:
-                self._Z = torch.normal(self.Z, noise, generator=self._random_state)
             loss[0] = self.lbfgs(max_iter=max_iter, **kwargs)
             if verbose:
-                print(f"LBFGS  {i+1:2d}: {loss[0]:.2f}")
-            if cc.has_converged(loss, self.copy):
-                break
-        if cc.optimal_state is not None:
-            self._Z = cc.optimal_state._Z
-            self._B = cc.optimal_state._B
-        return self.lbfgs(max_iter=max_iter * 2, **kwargs)
+                i += 1
+                print(f"LBFGS  {i:2d}: {loss[0]:.2f}")
+        self._Z = cc.optimal._Z
+        self._B = cc.optimal._B
+        loss = self.lbfgs(max_iter=max_iter * 2, **kwargs)
+        if verbose:
+            print(f"Final    : {loss:.2f}")
+        return loss
 
     optimize = optimise
 
@@ -796,6 +806,7 @@ class Slisemap:
         between: bool = True,
         escape_fn: Callable = escape_neighbourhood,
         loss: bool = False,
+        verbose: bool = False,
         **kwargs,
     ) -> Tuple[np.ndarray, np.ndarray, Optional[np.ndarray]]:
         """Generate embedding(s) and model(s) for new data item(s).
@@ -811,6 +822,7 @@ class Slisemap:
             between (bool, optional): If ``optimise=True``, should the new points affect each other? Defaults to True.
             escape_fn (Callable, optional): Escape function to use as initialisation (escape_neighbourhood/escape_greedy/escape_marginal). Defaults to escape_neighbourhood.
             loss (bool, optional): Return a vector of individual losses for the new items. Defaults to False.
+            verbose (bool, optional): Print status messages. Defaults to False.
             **kwargs: Optional keyword arguments to LBFGS.
 
         Returns:
@@ -819,6 +831,8 @@ class Slisemap:
         Xnew = self._as_new_X(Xnew)
         n = Xnew.shape[0]
         ynew = self._as_new_Y(ynew, n)
+        if verbose:
+            print("Escaping the new data")
         Bnew, Znew = escape_fn(
             X=Xnew,
             Y=ynew,
@@ -834,8 +848,13 @@ class Slisemap:
             Yold=self.Y,
             jit=self.jit,
         )
+        if verbose:
+            Zrad = torch.sqrt(torch.sum(Znew**2) / n).cpu().detach().item()
+            print("  radius(Z_new) =", Zrad)
 
         if optimise:
+            if verbose:
+                print("Optimising the new data")
             lf, set_new = make_marginal_loss(
                 X=self.X,
                 Y=self.Y,
@@ -864,11 +883,22 @@ class Slisemap:
                     LBFGS(lambda: lf(Bi, Zi), [Bi, Zi], **kwargs)
                     Bnew[j] = Bi.detach()
                     Znew[j] = Zi.detach()
+            if verbose:
+                Zrad = torch.sqrt(torch.sum(Znew**2) / n).cpu().detach().item()
+                print("  radius(Z_new) =", Zrad)
         if self.radius > 0:
-            Zout = Znew * (self.radius / (torch.sum(self._Z**2) + 1e-8))
+            if verbose:
+                print("Normalising the solution")
+            norm = self.radius / (torch.sqrt(torch.sum(self._Z**2) / self.n) + 1e-8)
+            Zout = Znew * norm
+            if verbose:
+                Zrad = torch.sqrt(torch.sum(Zout**2) / n).cpu().detach().item()
+                print("  radius(Z_new) =", Zrad)
         else:
             Zout = Znew
         if loss:
+            if verbose:
+                print("Calculating individual losses")
             lf = self.get_loss_fn(individual=True)
             if between:
                 loss = lf(
@@ -877,7 +907,7 @@ class Slisemap:
                     B=torch.cat((self.B, Bnew), 0),
                     Z=torch.cat((self.Z, Znew), 0),
                 )[self.n :]
-                loss = _tonp(loss)
+                loss = tonp(loss)
             else:
                 if self._jit:
                     lf = torch.jit.trace(lf, (Xnew[:1], ynew[:1], Bnew[:1], Znew[:1]))
@@ -890,9 +920,11 @@ class Slisemap:
                         Z=torch.cat((self.Z, Znew[None, j]), 0),
                     )[-1]
                     loss[j] = l.detach().cpu().item()
-            return _tonp(Bnew), _tonp(Zout), loss
+            if verbose:
+                print("  mean(loss) =", loss.mean())
+            return tonp(Bnew), tonp(Zout), loss
         else:
-            return _tonp(Bnew), _tonp(Zout)
+            return tonp(Bnew), tonp(Zout)
 
     def predict(
         self,
@@ -936,7 +968,7 @@ class Slisemap:
             + self.ridge * torch.sum(Bnew**2)
         )
         LBFGS(loss, [Bnew], **kwargs)
-        return _tonp(torch.diagonal(self.local_model(Xnew, Bnew), dim1=0, dim2=1).T)
+        return tonp(torch.diagonal(self.local_model(Xnew, Bnew), dim1=0, dim2=1).T)
 
     def copy(self) -> "Slisemap":
         """Make a copy of this Slisemap that references as much of the same torch-data as possible.
@@ -950,17 +982,26 @@ class Slisemap:
         other._Z = other._Z.clone().detach()
         return other
 
-    def save(self, f: Union[str, PathLike, BinaryIO], **kwargs):
+    def save(
+        self, f: Union[str, PathLike, BinaryIO], any_extension: bool = False, **kwargs
+    ):
         """Save the Slisemap object to a file.
         This method uses ``torch.save``, which uses ``pickle`` for the non-pytorch properties.
         This comes with the normal caveats like lambda-functions not being supported.
         However, the default pickle module can be overridden (see ``torch.save``).
-        The default ending for ``torch.save`` is ".pt" and by default the file is compressed.
+        The default file extension is ".sm" and by default the file is compressed.
 
         Args:
             f (Union[str, PathLike, BinaryIO]): Either a Path-like object or a (writable) File-like object.
+            any_extension (bool, optional): Do not check the file extension. Defaults to False.
             **kwargs: Parameters forwarded to ``torch.save``.
         """
+        if not any_extension and isinstance(f, (str, PathLike)):
+            if not str(f).endswith(".sm"):
+                _warn(
+                    "When saving Slisemap objects, consider using the '.sm' extension for consistency.",
+                    Slisemap.save,
+                )
         loss = self._loss
         prng = self._random_state
         try:
@@ -998,11 +1039,11 @@ class Slisemap:
         sm.random_state = sm._rs0
         return sm
 
-    def _cluster_models(
+    def get_model_clusters(
         self, clusters: int, B: Optional[np.ndarray] = None, random_state: int = 42
     ) -> Tuple[np.ndarray, np.ndarray]:
-        """Get cluster labels from k-means on the local model coefficients.
-        This function uses KMeans from scikit-learn.
+        """Cluster the local model coefficients using k-means (from scikit-learn).
+        This method (with a fixed random seed) is used for plotting Slisemap solutions.
 
         Args:
             clusters (int): Number of clusters.
@@ -1012,13 +1053,15 @@ class Slisemap:
         Returns:
             Tuple[np.ndarray, np.ndarray]: vector of cluster labels an matrix of cluster centers.
         """
-        from sklearn.cluster import KMeans
-
         km = KMeans(clusters, random_state=random_state).fit(
             B if B is not None else self.get_B()
         )
-        influence = np.abs(km.cluster_centers_)
-        influence = influence.max(0) + influence.mean(0)
+        # Sort according to value for the most influential coefficient
+        influence = (
+            km.cluster_centers_.var(0)
+            + np.abs(km.cluster_centers_).mean(0)
+            + np.abs(km.cluster_centers_).max(0)
+        )
         col = np.argmax(influence)
         ord = np.argsort(km.cluster_centers_[:, col])
         return np.argsort(ord)[km.labels_], km.cluster_centers_[ord]
@@ -1080,7 +1123,7 @@ class Slisemap:
                 f"The number of variable names ({len(variables)}) must match the number of coefficients ({B.shape[1]})",
             )
         if isinstance(clusters, int):
-            clusters, centers = self._cluster_models(clusters, B)
+            clusters, centers = self.get_model_clusters(clusters, B)
             if bars:
                 y = np.arange(B.shape[1]) if variables is None else variables
                 if not isinstance(bars, bool):
@@ -1106,10 +1149,13 @@ class Slisemap:
             B = B[np.argsort(Z[:, 0])]
             _assert(not bars, "`bar!=False` requires that `clusters` is an integer")
         if clusters is None:
-            L = self.value(individual=True)
-            sns.scatterplot(x=Z[:, 0], y=Z[:, 1], hue=L, palette="crest", ax=ax1)
+            if Z.shape[0] == self.Z.shape[0]:
+                L = tonp(torch.diagonal(self.get_L(numpy=False))).ravel()
+                sns.scatterplot(x=Z[:, 0], y=Z[:, 1], hue=L, palette="crest", ax=ax1)
+                ax1.legend(title="Fidelity")
+            else:
+                sns.scatterplot(x=Z[:, 0], y=Z[:, 1], palette="crest", ax=ax1)
             sns.heatmap(B, ax=ax2, center=0, cmap="RdBu", robust=True)
-            ax1.legend(title="Fidelity")
             if variables is not None:
                 ax2.set_xticks(np.arange(len(variables)) + 0.5)
                 ax2.set_xticklabels(variables, rotation=30)
@@ -1175,8 +1221,6 @@ class Slisemap:
         Returns:
             Optional[sns.FacetGrid]: Seaborn FacetGrid if show=False.
         """
-        import pandas as pd
-
         Z = Z if Z is not None else self.get_Z(rotate=True)
         if Z.shape[1] == 1:
             Z = np.concatenate((Z, np.zeros_like(Z)), 1)
@@ -1198,19 +1242,9 @@ class Slisemap:
             if isinstance(index, int):
                 index = [index]
             L = self.get_L()[:, index]
-        df = pd.concat(
-            [
-                pd.DataFrame(
-                    {
-                        "SLISEMAP 1": Z[:, 0],
-                        "SLISEMAP 2": Z[:, 1],
-                        "Fidelity": loss,
-                        "i": i,
-                    }
-                )
-                for i, loss in enumerate(L.T)
-            ],
-            ignore_index=True,
+        df = dict_concat(
+            {"SLISEMAP 1": Z[:, 0], "SLISEMAP 2": Z[:, 1], "Fidelity": loss, "i": i}
+            for i, loss in enumerate(L.T)
         )
         kwargs.setdefault("palette", "crest")
         kwargs.setdefault("kind", "scatter")
@@ -1237,26 +1271,27 @@ class Slisemap:
         w = 1 / col_wrap
         h = 1 / ((L.shape[1] - 1) // col_wrap + 1)
         if selection and index is not None:
+            size = plt.rcParams["lines.markersize"] ** 2 * 3
             for i, ax in zip(index, g.axes.ravel()):
-                ax.scatter(Z[i, 0], Z[i, 1], 100, "#9b52b4", "X")
+                ax.scatter(Z[i, 0], Z[i, 1], size, "#fd8431", "X")
             g.add_legend(
                 legend,
                 "Fidelity",
                 loc="lower center" if inside else "upper right",
-                bbox_to_anchor=(1 - w, h * 0.4, w * 0.9, h * 0.5) if inside else None,
+                bbox_to_anchor=(1 - w, h * 0.35, w * 0.9, h * 0.6) if inside else None,
             )
             g.add_legend(
-                {"": Line2D([], [], None, "None", "#9b52b4", "X", 8)},
+                {"": Line2D([], [], None, "None", "#fd8431", "X", 5)},
                 "Selected",
                 loc="upper center" if inside else "lower right",
-                bbox_to_anchor=(1 - w, 0.0, w * 0.9, h * 0.4) if inside else None,
+                bbox_to_anchor=(1 - w, h * 0.05, w * 0.9, h * 0.3) if inside else None,
             )
         else:
             g.add_legend(
                 legend,
                 "Fidelity",
                 loc="center" if inside else "center right",
-                bbox_to_anchor=(1 - w, 0, w * 0.9, h) if inside else None,
+                bbox_to_anchor=(1 - w, 0.05, w * 0.9, h * 0.9) if inside else None,
             )
         g.set_titles("")
         plt.suptitle(title)
@@ -1305,8 +1340,6 @@ class Slisemap:
         Returns:
             Optional[sns.FacetGrid]: Seaborn FacetGrid if show=False.
         """
-        import pandas as pd
-
         if X is None:
             X = self.get_X(intercept=False)
         if Y is None:
@@ -1324,16 +1357,13 @@ class Slisemap:
 
         if not scatter:
             if isinstance(clusters, int):
-                clusters, _ = self._cluster_models(clusters, B)
+                clusters, _ = self.get_model_clusters(clusters, B)
             elif clusters is None:
                 legend_inside = False
-            df = pd.concat(
-                [
-                    pd.DataFrame(dict(var=n, Value=XY[:, i], Cluster=clusters))
-                    for v, XY in [(targets, Y), (variables, X)]
-                    for i, n in enumerate(v)
-                ],
-                ignore_index=True,
+            df = dict_concat(
+                {"var": n, "Value": XY[:, i], "Cluster": clusters}
+                for v, XY in [(targets, Y), (variables, X)]
+                for i, n in enumerate(v)
             )
             if kwargs.setdefault("kind", "kde") == "kde":
                 kwargs.setdefault("bw_adjust", 0.75)
@@ -1361,20 +1391,15 @@ class Slisemap:
                 )
             if jitter > 0:
                 Z = np.random.normal(Z, jitter)
-            df = pd.concat(
-                [
-                    pd.DataFrame(
-                        {
-                            "var": n,
-                            "Value": XY[:, i],
-                            "SLISEMAP 1": Z[:, 0],
-                            "SLISEMAP 2": Z[:, 1],
-                        }
-                    )
-                    for v, XY in [(targets, Y), (variables, X)]
-                    for i, n in enumerate(v)
-                ],
-                ignore_index=True,
+            df = dict_concat(
+                {
+                    "var": n,
+                    "Value": XY[:, i],
+                    "SLISEMAP 1": Z[:, 0],
+                    "SLISEMAP 2": Z[:, 1],
+                }
+                for v, XY in [(targets, Y), (variables, X)]
+                for i, n in enumerate(v)
             )
             kwargs.setdefault("palette", "rocket")
             kwargs.setdefault("kind", "scatter")
