@@ -3,7 +3,7 @@ This module contains various useful functions.
 """
 
 from timeit import default_timer as timer
-from typing import Any, Callable, List, Optional, Sequence, Union
+from typing import Any, Callable, Dict, Iterator, List, Optional, Sequence, Union
 from warnings import warn
 
 import numpy as np
@@ -32,7 +32,7 @@ def _warn(warning: str, method: Optional[Callable] = None):
         warn(f"{method.__qualname__}: {warning}", SlisemapWarning, 2)
 
 
-def _tonp(x: torch.Tensor) -> np.ndarray:
+def tonp(x: torch.Tensor) -> np.ndarray:
     """Convert a torch.Tensor to a numpy.ndarray.
 
     Args:
@@ -44,12 +44,16 @@ def _tonp(x: torch.Tensor) -> np.ndarray:
     return x.cpu().detach().numpy()
 
 
+_tonp = tonp
+
+
 class CheckConvergence:
     """An object that tries to estimate when an optimisation has converged.
     Use it for, e.g., escape+optimisation cycles in Slisemap.
 
     Args:
-        patience (int, optional): How long should the optimisation continue without improvement. Defaults to 3.
+        patience (float, optional): How long should the optimisation continue without improvement. Defaults to 3.
+        max_iter (int, optional): The maximum number of iterations. Defaults to 2**30.
     """
 
     __slots__ = {
@@ -57,15 +61,19 @@ class CheckConvergence:
         "best": "Best loss value, so far.",
         "counter": "Number of steps since the best loss value.",
         "patience": "Number of steps allowed without improvement.",
-        "optimal_state": "Cache for storing the state that produced the best loss value.",
+        "optimal": "Cache for storing the state that produced the best loss value.",
+        "max_iter": "The maximum number of iterations.",
+        "iter": "The current number of iterations.",
     }
 
-    def __init__(self, patience: int = 3):
+    def __init__(self, patience: float = 3, max_iter=1 << 30):
         self.current = np.inf
-        self.best = np.inf
+        self.best = np.asarray(np.inf)
         self.counter = 0
-        self.patience = max(1e-8, patience)
-        self.optimal_state = None
+        self.patience = patience
+        self.optimal = None
+        self.max_iter = max_iter
+        self.iter = 0
 
     def has_converged(
         self,
@@ -84,21 +92,21 @@ class CheckConvergence:
         Returns:
             bool: True if the optimisation has converged.
         """
+        self.iter += 1
         loss = np.asarray(loss)
         if np.any(np.isnan(loss)):
             _warn("Loss is `nan`", CheckConvergence.has_converged)
             return True
         if np.any(loss < self.best):
-            self.best = np.minimum(loss, self.best)
-            self.current = loss
             self.counter = 0  # Reset the counter if a new best
             if store is not None and loss.item(0) < self.best.item(0):
-                self.optimal_state = store()
-            return False
-        # Increase the counter if no improvement
-        self.counter += np.mean(self.current <= loss)
+                self.optimal = store()
+            self.best = np.minimum(loss, self.best)
+        else:
+            # Increase the counter if no improvement
+            self.counter += np.mean(self.current <= loss)
         self.current = loss
-        return self.counter >= self.patience  # Has the patience run out
+        return self.counter >= self.patience or self.iter >= self.max_iter
 
 
 def LBFGS(
@@ -161,7 +169,7 @@ def LBFGS(
 
 
 def PCA_rotation(
-    X: torch.Tensor, components: int, full: bool = True, niter: int = 10
+    X: torch.Tensor, components: int = -1, full: bool = True, niter: int = 10
 ) -> torch.Tensor:
     """Calculate the rotation matrix from PCA.
 
@@ -169,7 +177,7 @@ def PCA_rotation(
 
     Args:
         X (torch.Tensor): The original matrix.
-        components (int): The number of components in the embedding.
+        components (int, optional): The maximum number of components in the embedding. Defaults to ``min(*X.shape)``.
         full (bool, optional): Use a full SVD for the PCA (slower). Defaults to True.
         niter (int, optional): The number of iterations when a randomised approach is used. Defaults to 10.
 
@@ -177,9 +185,9 @@ def PCA_rotation(
         torch.Tensor: Rotation matrix that turns the original matrix into the embedded space.
     """
     try:
-        components = min(*X.shape, components)
+        components = min(*X.shape, components) if components > 0 else min(*X.shape)
         if full:
-            return torch.linalg.svd(X, full_matrices=False)[2][:, :components]
+            return torch.linalg.svd(X, full_matrices=False)[2].T[:, :components]
         else:
             return torch.pca_lowrank(X, components, center=False, niter=niter)[2]
     except:
@@ -189,39 +197,6 @@ def PCA_rotation(
         return z
 
 
-def varimax(
-    X: torch.Tensor, gamma: float = 1.0, max_iter: float = 100, tolerance: float = 1e-8
-) -> torch.Tensor:
-    """Rotate a matrix using varimax (so that the first dimension has the largest variance).
-
-    Code adapted from: http://en.wikipedia.org/wiki/Talk:Varimax_rotation
-
-    Args:
-        X (torch.Tensor): Matrix
-        gamma (float, optional): Learning rate. Defaults to 1.0.
-        max_iter (float, optional): Maximum number of iterations. Defaults to 100.
-        tolerance (float, optional): Early stopping tolerance (relative). Defaults to 1e-8.
-
-    Returns:
-        torch.Tensor: Rotated matrix.
-    """
-    gamma /= X.shape[0]
-    eye = torch.eye(X.shape[1], dtype=X.dtype, device=X.device)
-    R = eye.clone()
-    d = 0.0
-    for _ in range(max_iter):
-        XR = X @ R
-        u, s, vh = torch.linalg.svd(
-            X.T @ (XR**3 - gamma * (XR @ (eye * (XR.T @ XR)))), full_matrices=False
-        )
-        R = u @ vh
-        d_old = d
-        d = torch.sum(s)
-        if d_old != 0.0 and d / d_old < 1.0 + tolerance:
-            break
-    return X @ R
-
-
 def global_model(
     X: torch.Tensor,
     Y: torch.Tensor,
@@ -258,3 +233,61 @@ def global_model(
 
     LBFGS(loss, [B])
     return B.detach()
+
+
+def dict_array(dict: Dict[str, Any]) -> Dict[str, np.ndarray]:
+    """Turn a dictionary of various values to a dictionary of numpy arrays with equal length inplace.
+
+    Args:
+        dict (Dict[str, Any]): Dictionary.
+
+    Returns:
+        Dict[str, np.ndarray]: The same dictionary where the values are numpy arrays with equal length.
+    """
+    n = 1
+    for k, v in dict.items():
+        v = np.asarray(v).ravel()
+        dict[k] = v
+        n = max(n, len(v))
+    for k, v in dict.items():
+        if len(v) == 1:
+            dict[k] = np.repeat(v, n)
+        elif len(v) != n:
+            _warn(f"Uneven lengths in dictionary ({k}: {len(v)} != {n})", dict_array)
+    return dict
+
+
+def dict_append(df: Dict[str, np.ndarray], d: Dict[str, Any]) -> Dict[str, np.ndarray]:
+    """Append a dictionary of values to a dictionary of numpy arrays (see ``dict_array``) inplace.
+
+    Args:
+        df (Dict[str, np.ndarray]): Dictionary of numpy arrays.
+        d (Dict[str, Any]): Dictionary to append.
+
+    Returns:
+        Dict[str, np.ndarray]: The same dictionary as ``df`` with the values from ``d`` appended.
+    """
+    d = dict_array(d)
+    for k in df.keys():
+        df[k] = np.concatenate((df[k], d[k]), 0)
+    return df
+
+
+def dict_concat(
+    dicts: Union[Sequence[Dict[str, Any]], Iterator[Dict[str, Any]]]
+) -> Dict[str, np.ndarray]:
+    """Combine multiple dictionaries into one by concatenating the values.
+    Calls ``dict_array`` to pre-process the dictionaries.
+
+    Args:
+        dicts (Union[Sequence[Dict[str, Any]], Iterator[Dict[str, Any]]]): Sequence or Generator with dictionaries (all must have the same keys).
+
+    Returns:
+        Dict[str, np.ndarray]: Combined dictionary.
+    """
+    if isinstance(dicts, Sequence):
+        dicts = (d for d in dicts)
+    df = dict_array(next(dicts))
+    for d in dicts:
+        dict_append(df, d)
+    return df
